# -- coding: utf8 --

r"""
EOS object for SAFT association sites contributions to the Helmholtz energy
"""

import sys
import numpy as np
import logging

from .compiled_modules.ext_Aassoc_numba import calc_Xika as calc_Xika_numba
from .compiled_modules.ext_Aassoc_python import calc_Xika as calc_Xika_python

logger = logging.getLogger(__name__)

if "cython" not in sys.modules:
    print("Cython package is unavailable, using Numba")
    flag_cython = True
else:
    flag_cython = False
    try:
        from .compiled_modules.ext_Aassoc_cython import calc_Xika as calc_Xika_cython
    except ImportError:
        raise ImportError(
            "Cython package is available but module: "
            "despasito.equations_of_state.saft.compiled_modules.ext_Aassoc_cython, has"
            " not been compiled."
        )


def _calc_Xika_wrap(*args, method_stat, maxiter=500, tol=1e-12, damp=0.1):
    r"""This function wrapper allows difference types of compiled functions to be "
    "referenced."""

    indices, rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc = args

    if len(np.shape(Kklab)) == 4:
        if method_stat.cython and flag_cython:
            Xika, _ = calc_Xika_cython(*args)
        elif method_stat.python:
            Xika, _ = calc_Xika_python(*args)
            logger.warning("Using pure python. Consider using 'numba' flag")
        elif method_stat.numba or not flag_cython:
            Xika, _ = calc_Xika_numba(*args)
        else:
            raise ValueError("Appropriate options for calc_Xika have not been defined.")

    elif len(np.shape(Kklab)) == 6:
        if method_stat.cython and flag_cython:
            Xika, _ = calc_Xika_cython(*args)
        elif method_stat.python:
            Xika, _ = calc_Xika_python(*args)
            logger.warning("Using pure python. Consider using 'numba' flag")
        elif method_stat.numba or not flag_cython:
            Xika, _ = calc_Xika_numba(*args)
        else:
            raise ValueError("Appropriate options for calc_Xika have not been defined.")

    return Xika


def assoc_site_indices(nk, molecular_composition, xi=None):
    r"""
    Make a list of sets of indices that allow quick identification of the relevant
    association sites.

    This is needed for solving Xika, the fraction of molecules of component i that are
    not bonded at a site of type a on group k.

    Parameters
    ----------
    nk : numpy.ndarray
        A matrix of (Nbeads x Nsites) Contains for each bead the number of each type
        of site
    molecular_composition : numpy.ndarray
        :math:`\nu_{i,k}/k_B`. Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    xi : numpy.ndarray, Optional, default=None
        Mole fraction of each component, sum(xi) should equal 1.0

    Returns
    -------
    indices : list[list]
        A list of sets of (component, bead, site) to identify the values of the Xika
        matrix that are being fit
    """

    indices = []

    # List of site indices for each bead type
    bead_sites = []
    for bead in nk:
        bead_sites.append([i for i, site in enumerate(bead) if site != 0])

    for i, comp in enumerate(molecular_composition):
        # if i not in zero_frac:
        for j, bead in enumerate(comp):
            if bead != 0 and bead_sites[j]:
                for k in bead_sites[j]:
                    indices.append([i, j, k])

    indices = np.array([np.array(x) for x in indices], dtype=int)

    return indices


def initiate_assoc_matrices(beads, bead_library, molecular_composition):
    r"""

    Generate matrices used for association site calculations.

    Compute epsilonHB (interaction energy for association term), Kklab (association
    interaction bonding volume, and nk (number of sites )

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - Nk\*: Optional, The number of sites of from list sitenames. Asterisk
        represents string from sitenames.

    molecular_composition : numpy.ndarray
        :math:`\nu_{i,k}/k_B`. Array of number of components by number of bead types.
        Defines the number of each type of group in each component.

    Returns
    ----------
    sitenames : list
        This list shows the names of the various association types found
    nk : numpy.ndarray
        A matrix of (Nbeads x Nsites) Contains for each bead the number of each type
        of site
    flag_assoc : bool
        If True, this flag indicates that association sites play a role in this
        system.
    """

    sitenames = []
    nk = [[] for i in range(len(beads))]

    for i, bead in enumerate(beads):
        nk[i] = [0 for x in sitenames]
        for key, value in bead_library[bead].items():
            if key.startswith("Nk"):
                tmp = key.split("-")
                if len(tmp) < 2:
                    raise ValueError("Association site names should be defined with hyphens " + "(e.g. Nk-H)")
                else:
                    _, site = tmp

                if site not in sitenames:
                    for j in range(i):
                        nk[j].append(0)
                    nk[i].append(value)
                    sitenames.append(site)
                else:
                    ind = sitenames.index(site)
                    nk[i][ind] = value
                logger.debug("Bead {} has {} of the association site {}".format(bead, value, site))

    indices = assoc_site_indices(nk, molecular_composition)
    if indices.size == 0:
        flag_assoc = False
    else:
        flag_assoc = True

    if flag_assoc:
        logger.info("The following association sites have been identified: {}".format(sitenames))
    else:
        logger.info("No association sites are used in this system.")

    return sitenames, np.array(nk), flag_assoc


def calc_assoc_matrices(
    beads,
    bead_library,
    molecular_composition,
    cross_library={},
    nk=None,
    sitenames=None,
):
    r"""

    Generate matrices used for association site calculations.

    Compute epsilonHB (interaction energy for association term), Kklab (association
    interaction bonding volume, and nk (number of sites )

    Note: Some papers use rc_klab instead of Kklab. In those cases, a function to
    calculate Kklab is needed (see Papaioannou 2014).

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - epsilonHB-\*-\*: Optional, Interaction energy between each bead and
        association site. Asterisk represents string from sitenames.
        - K-\*-\*: Optional, Bonding volume between each association site. Asterisk
        represents two strings from sitenames.
        - rc-\*-\*: Optional, Cutoff distance for association sites. Asterisk
        represents two strings from sitenames.
        - rd-\*-\*: Optional, Site position. Asterisk represents two strings from
        sitenames.
        - Nk-\*: Optional, The number of sites of from list sitenames. Asterisk
        represents string from sitenames.

    molecular_composition : numpy.ndarray
        :math:`\nu_{i,k}/k_B`. Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    cross_library : dict, Optional, default={}
        A dictionary where bead names are the keys to access a dictionary of a second
        tier of bead names. This structure contains the EOS cross interaction
        parameters:

        - epsilonHB-\*-\*: Optional, Interaction energy between each bead and
        association site. Asterisk represents string from sitenames.
        - K-\*-\*: Optional, Bonding volume between each association site. Asterisk
        represents two strings from sitenames.
        - rc-\*-\*: Optional, Cutoff distance for association sites. Asterisk
        represents two strings from sitenames.
        - rd-\*-\*: Optional, Site position. Asterisk represents two strings from
        sitenames.

    nk : numpy.ndarray, Optional, default=None
        A matrix of (Nbeads x Nsites) Contains for each bead the number of each type
        of site
    }
    sitenames : list, Optional, default=None
        This list shows the names of the various association types found

    Returns
    -------
    output_dict : dict
        This dictionary contains parameters relevant to calculating association site
        contributions. The following matrices may be inside, each of the size (ngroups,
        ngroups, nsites, nsites).

        - epsilonHB: Interaction energy between each bead and association site.
        - Kklab, Optional: Bonding volume between each association site
        - rc_klab, Optional: Cutoff distance for association sites
        - rd_klab, Optional: Association site position

    """

    nbeads = len(beads)
    if np.any(sitenames is None) or np.any(nk is None):
        sitenames, nk, _ = initiate_assoc_matrices(bead_library, beads, molecular_composition)
    else:
        nsitesmax = len(sitenames)
    epsilonHB = np.zeros((nbeads, nbeads, nsitesmax, nsitesmax))
    Kklab = np.zeros((nbeads, nbeads, nsitesmax, nsitesmax))
    rc_klab = np.zeros((nbeads, nbeads, nsitesmax, nsitesmax))
    rd_klab = np.zeros((nbeads, nbeads, nsitesmax, nsitesmax))

    flag_Kklab = False
    flag_rc_klab = False
    flag_rd_klab = False
    # self-interaction
    for i, nk1 in enumerate(nk):
        bead1 = beads[i]
        for a, site1 in enumerate(sitenames):

            if nk1[a] == 0.0:
                continue

            for b, site2 in zip(list(range(a, len(sitenames))), sitenames[a:]):

                if nk1[b] != 0:
                    epsilon_tmp = "-".join(["epsilonHB", site1, site2])
                    if epsilon_tmp not in bead_library[bead1]:
                        epsilon_tmp = "-".join(["epsilonHB", site2, site1])
                    K_tmp = "-".join(["K", site1, site2])
                    if K_tmp not in bead_library[bead1]:
                        K_tmp = "-".join(["K", site2, site1])
                    rc_tmp = "-".join(["rc", site1, site2])
                    if rc_tmp not in bead_library[bead1]:
                        rc_tmp = "-".join(["rc", site2, site1])
                    rd_tmp = "-".join(["rd", site1, site2])
                    if rd_tmp not in bead_library[bead1]:
                        rd_tmp = "-".join(["rd", site2, site1])

                    if epsilon_tmp in bead_library[bead1] and (
                        K_tmp not in bead_library[bead1] and rc_tmp not in bead_library[bead1]
                    ):
                        raise ValueError(
                            "An association site energy parameter for {}-{}".format(
                                site1,
                                site2,
                            )
                            + " was given for bead {}, but not the bonding".format(bead1)
                            + " information. Either K-{}-{}/K-{}-{} or".format(
                                site1,
                                site2,
                                site2,
                                site1,
                            )
                            + " rc-{}-{}/rc-{}-{} must be given.".format(site1, site2, site2, site1)
                        )
                    elif K_tmp in bead_library[bead1] and rc_tmp in bead_library[bead1]:
                        raise ValueError(
                            "Both association site bonding volumes and cutoff "
                            + "distances were provided for bead {}.".format(bead1)
                            + " This is redundant."
                        )
                    elif epsilon_tmp not in bead_library[bead1] and (
                        K_tmp in bead_library[bead1] or rc_tmp in bead_library[bead1]
                    ):
                        raise ValueError(
                            "An association site bonding information for {}".format("{}-{}".format(site1, site2))
                            + " was given for bead {}, but not the energy".format(bead1)
                            + " parameter. epsilonHB must be given."
                        )

                    if epsilon_tmp in bead_library[bead1]:
                        if a == b:
                            epsilonHB[i, i, a, b] = -1 * np.abs(bead_library[bead1][epsilon_tmp])
                        else:
                            epsilonHB[i, i, a, b] = bead_library[bead1][epsilon_tmp]
                        epsilonHB[i, i, b, a] = epsilonHB[i, i, a, b]
                    else:
                        continue

                    if K_tmp in bead_library[bead1]:
                        flag_Kklab = True
                        Kklab[i, i, a, b] = bead_library[bead1][K_tmp]
                        Kklab[i, i, b, a] = Kklab[i, i, a, b]

                    if rc_tmp in bead_library[bead1]:
                        flag_rc_klab = True
                        rc_klab[i, i, a, b] = bead_library[bead1][rc_tmp]
                        rc_klab[i, i, b, a] = rc_klab[i, i, a, b]

                    if rd_tmp in bead_library[bead1]:
                        flag_rd_klab = True
                        rd_klab[i, i, a, b] = bead_library[bead1][rd_tmp]
                        rd_klab[i, i, b, a] = rd_klab[i, i, a, b]

    # cross-interaction
    for i, nk1 in enumerate(nk):
        bead1 = beads[i]
        for a, site1 in enumerate(sitenames):
            if nk1[a] == 0.0:
                continue

            for b, site2 in enumerate(sitenames):
                epsilon_tmp = "-".join(["epsilonHB", site1, site2])
                K_tmp = "-".join(["K", site1, site2])
                rc_tmp = "-".join(["rc", site1, site2])
                rd_tmp = "-".join(["rd", site1, site2])
                for j, bead2 in enumerate(beads):
                    if i == j and a == b:
                        continue

                    if (
                        bead1 in cross_library
                        and bead2 in cross_library[bead1]
                        and epsilon_tmp in cross_library[bead1][bead2]
                    ):
                        # Update matrix if found in cross_library
                        if nk[i][a] == 0 or nk[j][b] == 0:
                            if 0 not in [nk[i][b], nk[j][a]]:
                                logger.warning(
                                    "Site names were listed in the wrong order for "
                                    + "parameter definitions in cross interaction "
                                    + "library. Changing {}_{} - {}_{}".format(
                                        beads[i],
                                        sitenames[a],
                                        beads[j],
                                        sitenames[b],
                                    )
                                    + " interaction to {}_{} - {}_{}".format(
                                        beads[i],
                                        sitenames[b],
                                        beads[j],
                                        sitenames[a],
                                    )
                                )
                                a, b = [b, a]
                            elif nk[i][a] == 0:
                                raise ValueError(
                                    "Cross interaction library parameters suggest a "
                                    + "{}_{} - {}_{} interaction, but {}".format(
                                        beads[i],
                                        sitenames[a],
                                        beads[j],
                                        sitenames[b],
                                        beads[i],
                                    )
                                    + " doesn't have site {}.".format(
                                        sitenames[a],
                                    )
                                )
                            elif nk[j][b] == 0:
                                raise ValueError(
                                    "Cross interaction library parameters suggest a "
                                    + "{}_{} - {}_{}   interaction, but {}".format(
                                        beads[i],
                                        sitenames[a],
                                        beads[j],
                                        sitenames[b],
                                        beads[j],
                                    )
                                    + " doesn't have site {}.".format(
                                        sitenames[b],
                                    )
                                )

                        epsilonHB[i, j, a, b] = cross_library[bead1][bead2][epsilon_tmp]
                        epsilonHB[j, i, b, a] = epsilonHB[i, j, a, b]

                        if flag_Kklab and K_tmp in cross_library[bead1][bead2]:
                            Kklab[i, j, a, b] = cross_library[bead1][bead2][K_tmp]
                            Kklab[j, i, b, a] = Kklab[i, j, a, b]

                        if flag_rc_klab and rc_tmp in cross_library[bead1][bead2]:
                            rc_klab[i, j, a, b] = cross_library[bead1][bead2][rc_tmp]
                            rc_klab[j, i, b, a] = rc_klab[i, j, a, b]

                        if flag_rd_klab and rd_tmp in cross_library[bead1][bead2]:
                            rd_klab[i, j, a, b] = cross_library[bead1][bead2][rd_tmp]
                            rd_klab[j, i, b, a] = rd_klab[i, j, a, b]

                    elif nk[j][b] != 0:
                        sitea = epsilon_tmp = "-".join(["epsilonHB", sitenames[a], sitenames[a]])
                        siteb = epsilon_tmp = "-".join(["epsilonHB", sitenames[b], sitenames[b]])
                        if (
                            epsilonHB[j, i, b, a] == 0.0
                            and sitea in bead_library[beads[i]]
                            and siteb in bead_library[beads[j]]
                        ):
                            epsilonHB[i, j, a, b] = np.sqrt(epsilonHB[i, i, a, a] * epsilonHB[j, j, b, b])
                            epsilonHB[i, j, a, b] *= -1 * np.sign(
                                bead_library[beads[i]][sitea] * bead_library[beads[j]][siteb]
                            )
                            epsilonHB[j, i, b, a] = epsilonHB[i, j, a, b]
                        if flag_Kklab and Kklab[i, j, a, b] == 0.0:
                            Kklab[i, j, a, b] = (
                                ((Kklab[i, i, a, a]) ** (1.0 / 3.0) + (Kklab[j, j, b, b]) ** (1.0 / 3.0)) / 2.0
                            ) ** 3
                            Kklab[j, i, b, a] = Kklab[i, j, a, b]

                        if flag_rc_klab and rc_klab[i, j, a, b] == 0.0:
                            rc_klab[i, j, a, b] = (rc_klab[i, i, a, a] + rc_klab[j, j, b, b]) / 2
                            rc_klab[j, i, b, a] = rc_klab[i, j, a, b]

                        if flag_rd_klab and rd_klab[i, j, a, b] == 0.0:
                            rd_klab[i, j, a, b] = (rd_klab[i, i, a, a] + rd_klab[j, j, b, b]) / 2
                            rd_klab[j, i, b, a] = rd_klab[i, j, a, b]

    output = {"epsilonHB": epsilonHB}
    if flag_Kklab:
        output["Kklab"] = Kklab
    if flag_rc_klab:
        output["rc_klab"] = rc_klab
    if flag_rd_klab:
        output["rd_klab"] = rd_klab

    if flag_Kklab and flag_rc_klab:
        raise ValueError(
            "Both association site bonding volumes and cutoff distances were provided." " This is redundant."
        )
    if flag_rd_klab and not flag_rc_klab:
        raise ValueError("Association site position were provided, but not cutoff distances.")

    return output


def calc_bonding_volume(rc_klab, dij_bar, rd_klab=None, reduction_ratio=0.25):
    """
    Calculate the association site bonding volume matrix

    Dimensions of (ncomp, ncomp, nbeads, nbeads, nsite, nsite)

    Parameters
    ----------
    rc_klab : numpy.ndarray
        This matrix of cutoff distances for association sites for each site type in
        each group type
    dij_bar : numpy.ndarray
        Component averaged hard sphere diameter
    rd_klab : numpy.ndarray, Optional, default=None
        Position of association site in each group (nbead, nbead, nsite, nsite)
    reduction_ratio : float, Optional, default=0.25
        Reduced distance of the sites from the center of the sphere of interaction.
        This value is used when site position, rd_klab is not defined for that
        site-site interaction.

    Returns
    -------
    Kijklab : numpy.ndarray
        Matrix of binding volumes
    """

    ncomp = len(dij_bar)
    nbead, _, nsite, _ = np.shape(rc_klab)
    Kijklab = np.zeros((ncomp, ncomp, nbead, nbead, nsite, nsite))

    for i in range(ncomp):
        for j in range(ncomp):
            for k in range(nbead):
                for l in range(nbead):
                    for a in range(nsite):
                        for b in range(nsite):
                            if rc_klab[k, l, a, b] != 0:
                                if np.all(rd_klab is None) or rd_klab[k, l, a, b] == 0:
                                    rd = reduction_ratio * dij_bar[i, j]
                                else:
                                    rd = rd_klab[k, l, a, b]

                                tmp0 = np.pi * dij_bar[i, j] ** 2 / (18 * rd**2)
                                tmp11 = np.log((rc_klab[k, l, a, b] + 2 * rd) / dij_bar[i, j])
                                tmp12 = 6 * rc_klab[k, l, a, b] ** 3 + 18 * rc_klab[k, l, a, b] ** 2 * rd - 24 * rd**3
                                tmp21 = rc_klab[k, l, a, b] + 2 * rd - dij_bar[i, j]
                                tmp22 = (
                                    22 * rd**2
                                    - 5 * rd * rc_klab[k, l, a, b]
                                    - 7 * rd * dij_bar[i, j]
                                    - 8 * rc_klab[k, l, a, b] ** 2
                                    + rc_klab[k, l, a, b] * dij_bar[i, j]
                                    + dij_bar[i, j] ** 2
                                )

                                Kijklab[i, j, k, l, a, b] = tmp0 * (tmp11 * tmp12 + tmp21 * tmp22)

    return Kijklab



r"""
Objects for storing and producing objective values for comparing experimental data to
EOS predictions.
"""

import numpy as np
import logging

from despasito.thermodynamics import thermo
from despasito.parameter_fitting import fit_functions as ff
from despasito.parameter_fitting.interface import ExpDataTemplate
import despasito.utils.general_toolbox as gtb
from despasito import fundamental_constants as constants

logger = logging.getLogger(__name__)


##################################################################
#                                                                #
#                              TLVE                              #
#                                                                #
##################################################################
class Data(ExpDataTemplate):
    r"""
    Object for Temperature dependent VLE data.

    This object is initiated in :func:`~despasito.parameter_fitting.fit` with the
    keyword, ``exp_data[*]["data_class_type"]="TLVE"``.

    This data could be evaluated with
    :func:`~despasito.thermodynamics.calculation_types.bubble_pressure` or
    :func:`~despasito.thermodynamics.calculation_types.dew_pressure`. Most entries in
    the exp. dictionary are converted to attributes.

    Parameters
    ----------
    data_dict : dict
        Dictionary of exp data of TLVE temperature dependent liquid vapor equilibria

        * calculation_type (str) - Optional, default='bubble_pressure', However,
        'dew_pressure' is also acceptable
        * MultiprocessingObject (obj) - Optional, Initiated
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
        * eos_obj (obj) - Equation of state object
        * T (list) - List of temperature values for calculation
        * P (list) - [Pa] List of pressure values for evaluation
        * xi(yi) (list) - List of liquid (or vapor) mole fractions used in
        bubble_pressure (or dew_pressure) calculation.
        * weights (dict) - A dictionary where each key is a system constraint
        (e.g. T or xi) which is also a header used in an optional exp. data file.
        The value associated with a header can be a list as long as the number of
        data points to multiply by the objective value associated with each point,
        or a float to multiply the objective value of this data set.
        * density_opts (dict) - Optional, default={}, Dictionary of options used in
        calculating pressure vs. mole fraction curves.
        * kwargs for
        :func:`~despasito.parameter_fitting.fit_functions.obj_function_form`

    Attributes
    ----------
    name : str
        Data type, in this case TLVE
    Eos : obj
        Equation of state object
    obj_opts : dict
        Keywords to compute the objective function with
        :func:`~despasito.parameter_fitting.fit_functions.obj_function_form`.
    npoints : int
        Number of sets of system conditions this object computes
    result_keys : list
        Thermodynamic property names used in calculation of objective function. In in
        this case: ["Plist", 'xilist'] or ["Plist", 'yilist']
    thermodict : dict
        Dictionary of inputs needed for thermodynamic calculations

        - calculation_type (str) default=bubble_pressure or dew_pressure
        - density_opts (dict) default={"min_density_fraction":(1.0 / 300000.0),
        "density_increment":10.0, "max_volume_increment":1.0E-4}

    weights : dict, Optional, default: {"some_property": 1.0 ...}
        Dictionary with keys corresponding to those in thermodict, with weighting
        factor or vector for each system property used in fitting

    """

    def __init__(self, data_dict):

        data_dict = data_dict.copy()

        super().__init__(data_dict)

        self.name = "TLVE"
        self.thermodict["density_opts"] = {}
        if "density_opts" in self.thermodict:
            self.thermodict["density_opts"].update(self.thermodict["density_opts"])

        if "T" in data_dict:
            self.thermodict["Tlist"] = data_dict["T"]
            del data_dict["T"]
        if "xi" in data_dict:
            self.thermodict["xilist"] = data_dict["xi"]
            del data_dict["xi"]
            if "xi" in self.weights:
                self.weights["xilist"] = self.weights.pop("xi")
        if "yi" in data_dict:
            self.thermodict["yilist"] = data_dict["yi"]
            del data_dict["yi"]
            if "yi" in self.weights:
                self.weights["yilist"] = self.weights.pop("yi")
        if "P" in data_dict:
            self.thermodict["Plist"] = data_dict["P"]
            self.thermodict["Pguess"] = data_dict["P"]
            del data_dict["P"]
            if "P" in self.weights:
                self.weights["Plist"] = self.weights.pop("P")

        self.thermodict.update(data_dict)

        thermo_keys = ["Plist", "xilist", "Tlist"]
        self.result_keys = ["Plist", "xilist", "yilist"]

        key_list = list(set(thermo_keys + self.result_keys))
        self.thermodict.update(gtb.check_length_dict(self.thermodict, key_list))
        self.npoints = np.size(self.thermodict["Tlist"])

        self.thermodict.update(
            gtb.set_defaults(
                self.thermodict,
                "Tlist",
                constants.standard_temperature,
                lx=self.npoints,
            )
        )

        self.weights.update(gtb.check_length_dict(self.weights, self.result_keys, lx=self.npoints))
        self.weights.update(gtb.set_defaults(self.weights, self.result_keys, 1.0))

        if "Tlist" not in self.thermodict:
            raise ImportError("Given TLVE data, values for T should have been provided.")

        thermo_keys = ["xilist", "yilist", "Plist"]
        if not any([key in self.thermodict for key in thermo_keys]):
            raise ImportError("Given TLVE data, mole fractions and/or pressure should have been " "provided.")

        if self.thermodict["calculation_type"] is None:
            if self.thermodict["xilist"]:
                self.thermodict["calculation_type"] = "bubble_pressure"
                logger.warning("No calculation type has been provided. Assume a calculation type " "of bubble_pressure")
            elif self.thermodict["yilist"]:
                self.thermodict["calculation_type"] = "dew_pressure"
                logger.warning("No calculation type has been provided. Assume a calculation type" " of dew_pressure")
            else:
                raise ValueError("Unknown calculation instructions")

        if self.thermodict["calculation_type"] == "bubble_pressure":
            self.result_keys.remove("xilist")
            del self.weights["xilist"]
        elif self.thermodict["calculation_type"] == "dew_pressure":
            self.result_keys.remove("yilist")
            del self.weights["yilist"]

        logger.info(
            "Data type 'TLVE' initiated with calculation_type, {}, and data "
            "types: {}.\nWeight data by: {}".format(
                self.thermodict["calculation_type"],
                ", ".join(self.result_keys),
                self.weights,
            )
        )

    def _thermo_wrapper(self):
        """
        Generate thermodynamic predictions from Eos object

        Returns
        -------
        phase_list : float
            A list of the predicted thermodynamic values estimated from thermo
            calculation. This list can be composed of lists or floats
        """

        # Remove results
        opts = self.thermodict.copy()
        tmp = self.result_keys + ["name", "parameters_guess"]
        for key in tmp:
            if key in opts:
                del opts[key]

        if self.thermodict["calculation_type"] == "bubble_pressure":
            try:
                output_dict = thermo(self.Eos, **opts)
                output = [output_dict["P"], output_dict["yi"]]
            except Exception:
                raise ValueError("Calculation of calc_bubble_pressure failed")

        elif self.thermodict["calculation_type"] == "dew_pressure":
            try:
                output_dict = thermo(self.Eos, **opts)
                output = [output_dict["P"], output_dict["xi"]]
            except Exception:
                raise ValueError("Calculation of calc_dew_pressure failed")

        return output

    def objective(self):
        """
        Generate objective function value from this dataset

        Returns
        -------
        obj_val : float
            A value for the objective function
        """

        # objective function
        phase_list = self._thermo_wrapper()
        phase_list, len_cluster = ff.reformat_output(phase_list)
        phase_list = np.transpose(np.array(phase_list))

        obj_value = np.zeros(2)

        if "Plist" in self.thermodict:
            obj_value[0] = ff.obj_function_form(
                phase_list[0], self.thermodict["Plist"], weights=self.weights["Plist"], **self.obj_opts
            )

        if self.thermodict["calculation_type"] == "bubble_pressure":
            if "yilist" in self.thermodict:
                yi = np.transpose(self.thermodict["yilist"])
                obj_value[1] = 0
                for i in range(len(yi)):
                    obj_value[1] += ff.obj_function_form(
                        phase_list[1 + i], yi[i], weights=self.weights["yilist"], **self.obj_opts
                    )
        elif self.thermodict["calculation_type"] == "dew_pressure":
            if "xilist" in self.thermodict:
                xi = np.transpose(self.thermodict["xilist"])
                obj_value[1] = 0
                for i in range(len(xi)):
                    obj_value[1] += ff.obj_function_form(
                        phase_list[1 + i], xi[i], weights=self.weights["xilist"], **self.obj_opts
                    )

        logger.info("Obj. breakdown for {}: P {}, zi {}".format(self.name, obj_value[0], obj_value[1]))

        if all([(np.isnan(x) or x == 0.0) for x in obj_value]):
            obj_total = np.inf
        else:
            obj_total = np.nansum(obj_value)

        return obj_total



"""
General functions that are applicable to multiple SAFT variants
"""

import numpy as np
import logging

from despasito.equations_of_state import constants

logger = logging.getLogger(__name__)


def calc_hard_sphere_matricies(T, sigmakl, bead_library, beads, Cprefactor_funcion):
    r"""
    Computes matrix of hard sphere interaction parameters dkk, dkl, and x0kl.

    This does not include function specific or association terms.

    Parameters
    ----------
    T : float
        Temperature of the system [K]
    sigmakl : numpy.ndarray
        Matrix of Mie diameter for groups (k,l)
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - epsilon: :math:`\epsilon_{k,k}/k_B`, Energy well depth scaled by Boltzmann
        constant
        - sigma: :math:`\sigma_{k,k}`, Size parameter [m]
        - mass: Bead mass [kg/mol]
        - lambdar: :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups
        of type k
        - lambdaa: :math:`\lambda^{a}_{k,k}`, Exponent of attractive term between
        groups of type k

    beads : list[str]
        List of unique bead names used among components
    Cprefactor_funcion : function
        Function used to calculate prefactor for potential

    Returns
    -------
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)
    x0kl : numpy.ndarray
        Matrix of sigmakl/dkl, sigmakl is the Mie radius for groups (k,l)
    """

    nbeads = np.size(beads)
    dkk = np.zeros(nbeads)
    for i in np.arange(nbeads):
        prefactor = Cprefactor_funcion(bead_library[beads[i]]["lambdar"], bead_library[beads[i]]["lambdaa"])
        dkk[i] = calc_dkk(
            bead_library[beads[i]]["epsilon"],
            bead_library[beads[i]]["sigma"],
            T,
            prefactor,
            bead_library[beads[i]]["lambdar"],
            bead_library[beads[i]]["lambdaa"],
        )
    dkl = np.zeros((nbeads, nbeads))
    for k in range(nbeads):
        for l in range(nbeads):
            dkl[k, l] = (dkk[k] + dkk[l]) / 2.0

    x0kl = sigmakl / dkl

    return dkl, x0kl


def _dkk_int(r, Ce_kT, sigma, lambdar, lambdaa):
    r"""
    Return integrand used to calculate the hard sphere diameter.

    :math:`d_{k,k}` of a group k. See eq. 10.

    Parameters
    ----------
    r : numpy.ndarray
        Bead distance between zero and :math:`sigma_{k,k}` [Å]
    Ce_kT : float
        :math:`C \epsilon_{k,k}/(k_B T)`, Mie prefactor scaled by kT
    sigma : float
        :math:`\sigma_{k,k}`, Size parameter [Å] (or same units as r)
    lambdar : float
        :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups of type k
    lambdaa : float
        :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups of type k

    Returns
    -------
    dkk_int_tmp : numpy.ndarray
        Integrand used to calculate the hard sphere diameter
    """

    dkk_int_tmp = 1.0 - np.exp(-Ce_kT * (np.power(sigma / r, lambdar) - np.power(sigma / r, lambdaa)))

    return dkk_int_tmp


def calc_dkk(epsilon, sigma, T, Cprefactor, lambdar, lambdaa=6.0):
    r"""
    Calculates hard sphere diameter of a group k, :math:`d_{k,k}`. Defined in eq. 10.
    using a 40 point Gauss Legendre

    Parameters
    ----------
    epsilon : float
        :math:`\epsilon_{k,k}/k_B`, Energy well depth scaled by Boltzmann constant [K]
    sigma : float
        :math:`\sigma_{k,k}`, Size parameter [Å] (or same units as r)
    T : float
        Temperature of the system [K]
    Cprefactor : float
        Prefactor for chosen potential
    lambdar : float
        :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups of type k
    lambdaa : float, Optional, default=6.0
        :math:`\lambda^{r}_{k,k}`, Exponent of repulsive term between groups of type k

    Returns
    -------
    dkk : float
        Hard sphere diameter of a group [Å]
    """

    Ce_kT = Cprefactor * epsilon / T
    # calculate integral of dkk_int from 0.0 to sigma

    w = np.array(
        [
            0.077505948,
            0.077505948,
            0.077039818,
            0.077039818,
            0.076110362,
            0.076110362,
            0.074723169,
            0.074723169,
            0.072886582,
            0.072886582,
            0.070611647,
            0.070611647,
            0.067912046,
            0.067912046,
            0.064804013,
            0.064804013,
            0.061306242,
            0.061306242,
            0.057439769,
            0.057439769,
            0.053227847,
            0.053227847,
            0.048695808,
            0.048695808,
            0.043870908,
            0.043870908,
            0.038782168,
            0.038782168,
            0.033460195,
            0.033460195,
            0.027937007,
            0.027937007,
            0.022245849,
            0.022245849,
            0.016421058,
            0.016421058,
            0.010498285,
            0.010498285,
            0.004521277,
            0.004521277,
        ]
    )
    x = np.array(
        [
            -0.038772418,
            0.038772418,
            -0.116084071,
            0.116084071,
            -0.192697581,
            0.192697581,
            -0.268152185,
            0.268152185,
            -0.341994091,
            0.341994091,
            -0.413779204,
            0.413779204,
            -0.483075802,
            0.483075802,
            -0.549467125,
            0.549467125,
            -0.61255389,
            0.61255389,
            -0.671956685,
            0.671956685,
            -0.727318255,
            0.727318255,
            -0.778305651,
            0.778305651,
            -0.824612231,
            0.824612231,
            -0.865959503,
            0.865959503,
            -0.902098807,
            0.902098807,
            -0.932812808,
            0.932812808,
            -0.957916819,
            0.957916819,
            -0.97725995,
            0.97725995,
            -0.990726239,
            0.990726239,
            -0.99823771,
            0.99823771,
        ]
    )

    r = 0.5 * sigma * (x + 1)
    dkk = 0.5 * sigma * np.sum(w * _dkk_int(r, Ce_kT, sigma, lambdar, lambdaa))

    return dkk


def calc_composition_dependent_variables(xi, molecular_composition, bead_library, beads):
    r"""
    Calculate the factor for converting molar density to bead density and molecular
    mole fractions to bead fractions.

    Parameters
    ----------
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    molecular_composition : numpy.array
        :math:`\nu_{i,k}/k_B`, Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
        Defined for eq. 11. Note that indices are flipped from definition in reference.
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters:

        - Vks: :math:`V_{k,s}`, Number of groups, k, in component
        - Sk: Optional, :math:`S_{k}`, Shape parameter of group k

    beads : list[str]
        List of unique bead names used among components

    Returns
    -------
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    xskl : numpy.ndarray
        Matrix of mole fractions of bead (i.e. segment or group) k multiplied by that
        of bead l
    """

    # compute Conversion factor
    Cmol2seg = 0.0
    for i in range(np.size(xi)):
        for j in range(np.size(beads)):
            Cmol2seg += (
                xi[i] * molecular_composition[i, j] * bead_library[beads[j]]["Vks"] * bead_library[beads[j]]["Sk"]
            )

    # initialize variables and arrays
    nbeads = len(beads)
    xsk = np.zeros(nbeads, float)
    # compute xsk
    for k in range(nbeads):
        xsk[k] = np.sum(xi * molecular_composition[:, k]) * bead_library[beads[k]]["Vks"] * bead_library[beads[k]]["Sk"]
    xsk /= Cmol2seg

    # calculate  xskl matrix
    xskl = np.zeros((nbeads, nbeads))
    for k in range(nbeads):
        for l in range(nbeads):
            xskl[k, l] = xsk[k] * xsk[l]

    return Cmol2seg, xskl


def calc_zetaxstar(rho, Cmol2seg, xskl, sigmakl):
    r"""
    Calculate matrix of hypothetical packing fraction based on sigma for groups (k,l)

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [:math:`mol/m^3`]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    xskl : numpy.ndarray
        Matrix of mole fractions of bead (i.e. segment or group) k multiplied by
        bead l
    sigmakl : numpy.ndarray
        Matrix of Mie diameter for groups (k,l)

    Returns
    -------
    zetaxstar : numpy.ndarray
        Matrix of hypothetical packing fraction based on sigma for groups (k,l)
    """

    # compute zetaxstar eq. 35
    zetaxstar = rho * Cmol2seg * ((np.pi / 6.0) * np.sum(xskl * (sigmakl**3 * constants.molecule_per_nm3)))

    return zetaxstar


def calc_zetax(rho, Cmol2seg, xskl, dkl):
    r"""
    Calculate matrix of hypothetical packing fraction based on hard sphere diameter for
    groups (k,l)

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [:math:`mol/m^3`]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    xskl : numpy.ndarray
        Matrix of mole fractions of bead (i.e. segment or group) k multiplied by bead l
    dkl : numpy.ndarray
        Matrix of hardsphere diameters for groups (k,l)

    Returns
    -------
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)
    """
    zetax = rho * Cmol2seg * ((np.pi / 6.0) * np.sum(xskl * (dkl**3 * constants.molecule_per_nm3)))

    return zetax


def calc_KHS(zetax):
    r"""
    Calculate (length of density array) isothermal compressibility of system with
    packing fraction zetax

    Parameters
    ----------
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)

    Returns
    -------
    KHS : numpy.ndarray
        (length of densities) isothermal compressibility of system with packing
        fraction zetax
    """
    KHS = ((1.0 - zetax) ** 4) / (1.0 + (4.0 * zetax) + (4.0 * (zetax**2)) - (4.0 * (zetax**3)) + (zetax**4))

    return KHS



# -- coding: utf8 --

import numpy as np
import logging

from despasito.equations_of_state import constants

logger = logging.getLogger(__name__)


def calc_Xika(
    indices,
    rho,
    xi,
    molecular_composition,
    nk,
    Fklab,
    Kklab,
    gr_assoc,
    maxiter=500,
    tol=1e-12,
    damp=0.1,
):
    r"""
    Calculate the fraction of molecules of component i that are not bonded at a site
    of type a on group k.

    Parameters
    ----------
    indices : list[list]
        A list of sets of (component, bead, site) to identify the values of the Xika
        matrix that are being fit
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    molecular_composition : numpy.array
        :math:`\nu_{i,k}/k_B`, Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    nk : numpy.ndarray
        For each bead the number of each type of site
    Fklab : numpy.ndarray
        The association strength between a site of type a on a group of type k of
        component i and a site of type b on a group of type l of component j., known
        as the Mayer f-function.
    gr_assoc : numpy.ndarray
        Reference fluid pair correlation function used in calculating association
        sites, (len(rho) x Ncomp x Ncomp)
    maxiter : int, Optional, default=500
        Maximum number of iteration for minimization
    tol : float, Optional, default=1e-12
        Once matrix converges.
    damp : float, Optional, default=0.1
        Only add a fraction of the new matrix values to update the guess

    Returns
    -------
    Xika : numpy.ndarray
        The fraction of molecules of component i that are not bonded at a site of type
        a on group k. Matrix (len(rho) x Ncomp x Nbeads x len(sitenames))
    err_array : numpy.ndarray
        Of the same length of rho, is a list in the error of the total error Xika for
        each point.
    """

    # ncomp, nbeads = np.shape(molecular_composition)
    # nsitesmax = np.shape(nk)[1]
    nrho = len(rho)
    l_ind = len(indices)
    l_K = len(np.shape(Kklab))

    #    Xika_final = np.ones((nrho,ncomp, nbeads, nsitesmax))
    Xika_final = np.ones((nrho, len(indices)))
    err_array = np.zeros(nrho)

    # Parallelize here, with respect to rho!
    Xika_elements_old = 0.5 * np.ones(len(indices))
    for r in range(nrho):
        for knd in range(maxiter):
            Xika_elements_new = np.ones(len(Xika_elements_old))
            for ind in range(l_ind):
                i, k, a = indices[ind]
                for jnd in range(l_ind):
                    j, l, b = indices[jnd]
                    if l_K == 4:
                        delta = Fklab[k, l, a, b] * Kklab[k, l, a, b] * gr_assoc[r, i, j]
                    elif l_K == 6:
                        delta = Fklab[k, l, a, b] * Kklab[i, j, k, l, a, b] * gr_assoc[r, i, j]

                    Xika_elements_new[ind] += (
                        constants.molecule_per_nm3
                        * rho[r]
                        * xi[j]
                        * molecular_composition[j, l]
                        * nk[l, b]
                        * Xika_elements_old[jnd]
                        * delta
                    )

            Xika_elements_new = 1.0 / Xika_elements_new
            obj = np.sum(np.abs(Xika_elements_new - Xika_elements_old))

            if obj < tol:
                break
            else:
                if obj / max(Xika_elements_old) > 1e3:
                    Xika_elements_old = Xika_elements_old + damp * (Xika_elements_new - Xika_elements_old)
                else:
                    Xika_elements_old = Xika_elements_new

        err_array[r] = obj

        Xika_final[r, :] = Xika_elements_old
        # for jjnd in range(l_ind):
        #    i,k,a = indices[jjnd]
        #    Xika_final[r,i,k,a] = Xika_elements[jjnd]

    return Xika_final, err_array



""" Interface needed to create further objects to represent experimental data.
"""

# All folders in this directory refer back to this interface
import numpy as np
import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


class ExpDataTemplate(ABC):
    r"""
    Interface needed to create further objects to represent experimental data.

    Parameters
    ----------
    data_dict : dict
        Dictionary of exp data

        * calculation_type (str) - Optional, default=*to be set*
        * MultiprocessingObject (obj) - Optional, Initiated
        :class:`~despasito.utils.parallelization.MultiprocessingJob`
        * eos_obj (obj) - Equation of state object
        * weights (dict) - A dictionary where each key is the header used in the exp.
        data file. The value associated with a header can be a list as long as the
        number of data points to multiply by the objective value associated with each
        point, or a float to multiply the objective value of this data set.
        * density_opts (dict) - Optional, default={}, Dictionary of options used in
        calculating pressure vs. mole fraction curves.
        * Allowed property keys and associated values
        * kwargs for
        :func:`~despasito.parameter_fitting.fit_functions.obj_function_form`

    Attributes
    ----------
    name : str
        Data type, name of module, "To be set"
    Eos : obj
        Equation of state object
    weights : dict, Optional, default: {"some_property": 1.0 ...}
        Dictionary corresponding to thermo_dict, with weighting factor or vector for
        each system property used in fitting
    obj_opts : dict
        Keywords to compute the objective function with
        :func:`~despasito.parameter_fitting.fit_functions.obj_function_form`.
    npoints : int
        Number of sets of system conditions this object computes
    thermodict : dict
        Dictionary of inputs needed for thermodynamic calculations

        - calculation_type (str) default=*to be set*
        - density_opts (dict) default={}

    """

    def __init__(self, data_dict):

        # Self interaction parameters
        self.name = "To be set"

        if "eos_obj" in data_dict:
            self.Eos = data_dict["eos_obj"]
            del data_dict["eos_obj"]
        else:
            raise ValueError("An Eos object should have been included")

        if "weights" in data_dict:
            self.weights = data_dict["weights"]
            del data_dict["weights"]
        else:
            self.weights = {}

        self.obj_opts = {}
        if "objective_method" in data_dict:
            self.obj_opts["method"] = data_dict["objective_method"]
            del data_dict["objective_method"]
        fitting_opts = ["nan_number", "nan_ratio"]
        for key in fitting_opts:
            if key in data_dict:
                self.obj_opts[key] = data_dict[key]
                del data_dict[key]
        logger.info("Objective function options: {}".format(self.obj_opts))

        self.npoints = np.nan

        # Add to thermo_dict
        self.thermodict = {"calculation_type": None}
        thermo_dict_keys = ["MultiprocessingObject", "density_opts", "calculation_type"]
        for key in thermo_dict_keys:
            if key in data_dict:
                self.thermodict[key] = data_dict[key]
                del data_dict[key]

    def update_parameters(self, fit_bead, param_names, param_values):
        r"""
        Update a single parameter value during parameter fitting process.

        To refresh those parameters that are dependent on to bead_library or
        cross_library, use method "parameter refresh".

        Parameters
        ----------
        fit_bead : str
            Name of bead being fit
        param_names : list
            Parameters to be fit. See EOS documentation for supported parameter names.
            Cross interaction parameter names should be composed of parameter name and
            the other bead type, separated by an underscore (e.g. epsilon_CO2).
        param_values : list
            Value of parameter

        """

        for i, param in enumerate(param_names):
            bead_names = [fit_bead]
            fit_parameter_names_list = param.split("_")
            param = fit_parameter_names_list[0]
            if len(fit_parameter_names_list) > 1:
                bead_names.append(fit_parameter_names_list[1])

            if len(fit_parameter_names_list) == 1:
                self.Eos.update_parameter(fit_parameter_names_list[0], [fit_bead], param_values[i])
            elif len(fit_parameter_names_list) == 2:
                self.Eos.update_parameter(
                    fit_parameter_names_list[0],
                    [fit_bead, fit_parameter_names_list[1]],
                    param_values[i],
                )
            else:
                raise ValueError(
                    "Parameters for only one bead are allowed to be fit. Multiple "
                    "underscores in a parameter name suggest more than one bead type "
                    "in your fit parameter name, {}".format(param)
                )

        if hasattr(self.Eos, "parameter_refresh"):
            self.Eos.parameter_refresh()

    @abstractmethod
    def objective(self):
        """Float representing objective function of from comparing predictions to
        experimental data."""
        pass

    def __str__(self):

        string = "Data Set Object\nName: {}\nCalculation_type: {}\nNumber of " "Points: {}".format(
            self.name, self.thermodict["calculation_type"], self.npoints
        )
        return string



"""
DESPASITO
DESPASITO: Determining Equilibrium State and Parametrization Application for SAFT, Intended for Thermodynamic Output
"""

import os
import glob
from setuptools import Extension, setup

import numpy as np
from Cython.Build import cythonize

fpath = os.path.join("despasito", "equations_of_state", "saft", "compiled_modules")
extensions = []

cython_list = glob.glob(os.path.join(fpath, "*.pyx"))
for cyext in cython_list:
    name = os.path.split(cyext)[-1].split(".")[-2]
    cy_ext_1 = Extension(
        name=os.path.join(fpath,name).replace(os.sep,"."), 
        sources=[cyext], 
        include_dirs=[fpath, np.get_include()],
        )
    extensions.extend(
        cythonize(
            [cy_ext_1],
            compiler_directives={
                'language_level': 3,
                'cdivision': False,
                "boundscheck": True
            }))

if __name__ == "__main__":
    setup(
        name='despasito',
        ext_modules=extensions,
    )



"""
Thermodynamics

This package will take in an equation of state object, and any user defined variables
for calculation. The calculation type will then be compared to those available in the
thermo.py file and be executed.

"""

from inspect import getmembers, isfunction

# import logging

from . import calculation_types

# logger = logging.getLogger(__name__)


def thermo(Eos, calculation_type=None, **kwargs):
    """
    Use factory design pattern to search for matching calculation_type with those
    supported in this module.

    To add a new calculation type, add a new wrapper function to
    ``calculation_types.py``.

    Parameters
    ----------
    Eos : obj
        Equation of state object with the following methods: ``pressure``,
        ``density_max``, and ``fugacity_coefficient``.
    calculation_type : str
        Calculation type supported in
        :mod:`~despasito.thermodynamics.calculation_types`
    kwargs
        Other keywords passed to the function, depends on calculation type

    Returns
    -------
    output_dict : dict
        Output of dictionary containing given and calculated values
    """

    if calculation_type is None:
        raise ValueError("No calculation type specified")

    # Extract available calculation types
    calc_list = [o[0] for o in getmembers(calculation_types) if isfunction(o[1])]

    # Unpack inputs and check
    try:
        func = getattr(calculation_types, calculation_type)

    except Exception:
        raise ImportError(
            "The calculation type, '{}', was not found\nThe following calculation "
            "types are supported: {}".format(calculation_type, ", ".join(calc_list))
        )

    try:
        output_dict = func(Eos, **kwargs)
    except Exception:
        raise TypeError("The calculation type, '{}', failed".format(calculation_type))

    return output_dict



"""
Regression tests for the despasito input/output package.

The following functions are not tested because they require importing a file:
    file2paramdict, extract_calc_data, process_commandline

The following functions are not tested because they require exporting a file:
    write_EOSparameters

The following functions are tested:
    process_bead_data
"""

# Import package, test suite, and other packages as needed
import despasito.input_output.read_input as ri
import pytest
import numpy as np


# Not Used because we shouldn't reference an external file
@pytest.mark.parametrize("data", [([[["CH4_2", 1]], [["eCH3", 2]]])])
def test_process_bead_data(data):
    """Test extraction of system component information"""

    beads, molecular_composition = ri.process_bead_data(data)

    errors = []
    if not set(beads) == set(["CH4_2", "eCH3"]):
        errors.append("beads: %s is not %s" % (str(beads), str(["CH4_2", "eCH3"])))
    elif not np.array_equal(molecular_composition, np.array([[1.0, 0.0], [0.0, 2.0]])):
        errors.append(
            "molecular_composition: %s is not %s"
            % (str(molecular_composition), str(np.array([[1.0, 0.0], [0.0, 2.0]])))
        )

    assert not errors, "errors occured:\n{}".format("\n".join(errors))



import numpy as np
from despasito.equations_of_state import initiate_eos

beads = ["CH3", "CH2"]
beads_per_molecule = np.array([[2.0, 8.0]])
bead_library = {
    "CH3": {
        "epsilon": 256.7662,
        "lambdaa": 6.0,
        "lambdar": 15.04982,
        "sigma": 4.077257e-1,
        "Sk": 0.5725512,
        "Vks": 1,
        "mass": 0.015035,
    },
    "CH2": {
        "epsilon": 473.3893,
        "lambdaa": 6.0,
        "lambdar": 19.87107,
        "sigma": 4.880081e-1,
        "Sk": 0.2293202,
        "Vks": 1,
        "mass": 0.014027,
    },
}
cross_library = {"CH3": {"CH2": {"epsilon": 350.770}}}
Eos = initiate_eos(
    eos="saft.gamma_mie",
    beads=beads,
    molecular_composition=beads_per_molecule,
    bead_library=bead_library,
    cross_library=cross_library,
    numba=True,
)

rho, T, xi = 553.0, 700.0, np.array([1.0])
AHS = Eos.saft_source.Ahard_sphere(rho, T, xi)
A1 = Eos.saft_source.Afirst_order(rho, T, xi)
A2 = Eos.saft_source.Asecond_order(rho, T, xi)
A3 = Eos.saft_source.Athird_order(rho, T, xi)
Am1 = AHS + A1 + A2 + A3
Am2 = Eos.Amonomer(rho, T, xi)
print(
    "The Monomer Contribution for Helmholtz Energy: {},\n".format(Am2)
    + "    equals the sum of its components: {}".format(Am1)
)



"""
    Interface needed to create further equation of state (EOS) objects.

    All folders in this directory refer back to this interface. Using this template
    all future EOS will be easily exchanged.
"""

# All folders in this directory refer back to this interface

import numpy as np
import logging
from abc import ABC, abstractmethod

logger = logging.getLogger(__name__)


# __________________ EOS Interface _________________
class EosTemplate(ABC):
    """
    Interface used in all EOS object options.

    By using this template, all EOS objects are then easily exchanged.

    Parameters
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters
    cross_library : dict, Optional
        Optional library of bead cross interaction parameters. As many or as few of
        the desired parameters may be defined for whichever group combinations are
        desired. The remaining are estimated with mixing rules.
    kwargs
        Additional keywords from EOS object type

    Attributes
    ----------
    beads : list[str]
        List of unique bead names used among components
    bead_library : dict
        A dictionary where bead names are the keys to access EOS self interaction
        parameters
    cross_library : dict
        Library of bead cross interaction parameters. As many or as few of the desired
        parameters may be defined for whichever group combinations are desired.
    parameter_types : list[str]
        This list of parameter names must be defined in an EOS object for parameter
        fitting
    parameter_bound_extreme : dict
        With each parameter names as an entry representing a list with the minimum and
        maximum feasible parameter value.
    number_of_components : int
        Number of components in mixture represented by given EOS object.

    """

    def __init__(self, beads, bead_library, **kwargs):
        """Initiation of EOS object with attributes needed by other modules."""

        self.parameter_types = None
        self.parameter_bound_extreme = None

        self.number_of_components = None
        for bead in beads:
            if bead not in bead_library:
                raise ValueError("The group, '{}', was not found in parameter library".format(bead))

        self.beads = None
        self.bead_library = None
        self.cross_library = None

    @abstractmethod
    def pressure(self, rho, T, xi):
        """
        Compute pressure given system information

        Parameters
        ----------
        rho : numpy.ndarray
            Number density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component

        Returns
        -------
        P : numpy.ndarray
            Array of pressure values [Pa] associated with each density and so equal in
            length
        """
        pass

    @abstractmethod
    def fugacity_coefficient(self, P, rho, xi, T):
        r"""
        Compute fugacity coefficient

        Parameters
        ----------
        P : float
            Pressure of the system [Pa]
        rho : float
            Molar density of system [mol/m^3]
        T : float
            Temperature of the system [K]
        xi : list[float]
            Mole fraction of each component

        Returns
        -------
        fugacity_coefficient : numpy.ndarray
            :math:`\mu_i`, Array of fugacity coefficient values for each component
        """
        pass

    @abstractmethod
    def density_max(self, xi, T, maxpack=0.9):
        """
        Estimate the maximum density based on the hard sphere packing fraction.

        Parameters
        ----------
        xi : list[float]
            Mole fraction of each component
        T : float
            Temperature of the system [K]
        maxpack : float, Optional, default=0.9
            Maximum packing fraction

        Returns
        -------
        max_density : float
            Maximum molar density [mol/m^3]
        """
        pass

    def guess_parameters(self, param_name, bead_names):
        """
        Output a guess for the given parameter type.
        """
        """
        Generate initial guesses for the parameters to be fit.

        Parameters
        ----------
        parameter : str
            Parameter to be fit. See EOS documentation for supported parameter names.
        bead_names : list
            Bead names to be changed. For a self interaction parameter, the length will
            be 1, for a cross interaction parameter, the length will be two.

        Returns
        -------
        param_initial_guess : numpy.ndarray,
            An initial guess for parameter, it will be optimized throughout the p
            rocess.
        """

        keys = ["beads", "parameter_types", "bead_library", "cross_library"]
        for key in keys:
            if getattr(self, key) is None:
                raise ValueError(
                    "EOS object attribute, {}, cannot be None. Ensure EOS object "
                    "initiates this attribute".format(key)
                )

        if len(bead_names) > 2:
            raise ValueError(
                "The bead names {} were given, but only a maximum of 2 are " "permitted.".format(", ".join(bead_names))
            )
        if not set(bead_names).issubset(self.beads):
            raise ValueError(
                "The bead names {} were given, but they are not in the allowed"
                " list: {}".format(", ".join(bead_names), ", ".join(self.beads))
            )

        param_value = None
        # Self interaction parameter
        if len(bead_names) == 1:
            if bead_names[0] in self.bead_library and param_name in self.bead_library[bead_names[0]]:
                param_value = self.bead_library[bead_names[0]][param_name]
        # Cross interaction parameter
        elif len(bead_names) == 2:
            if bead_names[1] in self.cross_library and bead_names[0] in self.cross_library[bead_names[1]]:
                if param_name in self.cross_library[bead_names[1]][bead_names[0]]:
                    param_value = self.cross_library[bead_names[1]][bead_names[0]][param_name]
            elif bead_names[0] in self.cross_library and bead_names[1] in self.cross_library[bead_names[0]]:
                if param_name in self.cross_library[bead_names[0]][bead_names[1]]:
                    param_value = self.cross_library[bead_names[0]][bead_names[1]][param_name]

        if param_value is None:
            bounds = self.check_bounds(bead_names[0], param_name, np.empty(2))
            param_value = (bounds[1] - bounds[0]) / 2 + bounds[0]

        return param_value

    def check_bounds(self, parameter, param_name, bounds):
        """
        Generate initial guesses for the parameters to be fit.

        Parameters
        ----------
        parameter : str
            Parameter to be fit. See EOS documentation for supported parameter names.
        param_name : str
            Full parameter string to be fit. See EOS documentation for supported
            parameter names.
        bounds : list
            Upper and lower bound for given parameter type

        Returns
        -------
        bounds : list
            A screened and possibly corrected low and a high value for the parameter,
            param_name
        """

        keys = ["parameter_types", "parameter_bound_extreme"]
        for key in keys:
            if getattr(self, key) is None:
                raise ValueError(
                    "EOS object attribute, {}, cannot be None. Ensure EOS object "
                    "initiates this attribute".format(key)
                )

        fit_parameter_names_list = param_name.split("_")
        parameter = fit_parameter_names_list[0]

        bounds_new = np.zeros(2)
        # Non bonded parameters
        if parameter in self.parameter_bound_extreme:

            if bounds[0] < self.parameter_bound_extreme[parameter][0]:
                logger.debug(
                    "Given {} lower boundary, {}, is less than what is recommended by "
                    "Eos object. Using value of {}.".format(
                        param_name,
                        bounds[0],
                        self.parameter_bound_extreme[parameter][0],
                    )
                )
                bounds_new[0] = self.parameter_bound_extreme[parameter][0]
            elif bounds[0] > self.parameter_bound_extreme[parameter][1]:
                logger.debug(
                    "Given {} lower boundary, {}, is greater than what is recommended "
                    "by Eos object. Using value of {}.".format(
                        param_name,
                        bounds[0],
                        self.parameter_bound_extreme[parameter][0],
                    )
                )
                bounds_new[0] = self.parameter_bound_extreme[parameter][0]
            else:
                bounds_new[0] = bounds[0]

            if bounds[1] > self.parameter_bound_extreme[parameter][1]:
                logger.debug(
                    "Given {} upper boundary, {}, is greater than what is recommended "
                    "by Eos object. Using value of {}.".format(
                        param_name,
                        bounds[1],
                        self.parameter_bound_extreme[parameter][1],
                    )
                )
                bounds_new[1] = self.parameter_bound_extreme[parameter][1]
            elif bounds[1] < self.parameter_bound_extreme[parameter][0]:
                logger.debug(
                    "Given {} upper boundary, {}, is less than what is recommended "
                    "by Eos object. Using value of {}.".format(
                        param_name,
                        bounds[1],
                        self.parameter_bound_extreme[parameter][1],
                    )
                )
                bounds_new[1] = self.parameter_bound_extreme[parameter][1]
            else:
                bounds_new[1] = bounds[1]

        else:
            raise ValueError(
                "The parameter name {} is not found in the allowed parameter types:"
                " {}".format(param_name, ", ".join(self.parameter_types))
            )

        return bounds_new

    def update_parameter(self, param_name, bead_names, param_value):
        r"""
        Update a single parameter value during parameter fitting process.

        Parameters
        ----------
        param_name : str
            Parameter to be fit. See EOS documentation for supported parameter names.
            Cross interaction parameter names should be composed of parameter name and
            the other bead type, separated by an underscore (e.g. epsilon_CO2).
        bead_names : list
            Bead names to be changed. For a self interaction parameter, the length will
            be 1, for a cross interaction parameter, the length will be two.
        param_value : float
            Value of parameter

        """

        keys = ["beads", "parameter_types", "bead_library", "cross_library"]
        for key in keys:
            if getattr(self, key) is None:
                raise ValueError(
                    "EOS object attribute, {}, cannot be None. Ensure EOS object "
                    "initiates this attribute".format(key)
                )

        if len(bead_names) > 2:
            raise ValueError(
                "The bead names {} were given, but only a maximum of 2 are " "permitted.".format(", ".join(bead_names))
            )
        if not set(bead_names).issubset(self.beads):
            raise ValueError(
                "The bead names {} were given, but they are not in the allowed "
                "list: {}".format(", ".join(bead_names), ", ".join(self.beads))
            )

        if not any([x in param_name for x in self.parameter_types]):
            raise ValueError(
                "The parameter name {} is not found in the allowed parameter "
                "types: {}".format(param_name, ", ".join(self.parameter_types))
            )

        # Self interaction parameter
        if len(bead_names) == 1:
            if bead_names[0] in self.bead_library:
                self.bead_library[bead_names[0]][param_name] = param_value
            else:
                self.bead_library[bead_names[0]] = {param_name: param_value}
        # Cross interaction parameter
        elif len(bead_names) == 2:
            if bead_names[1] in self.cross_library and bead_names[0] in self.cross_library[bead_names[1]]:
                self.cross_library[bead_names[1]][bead_names[0]][param_name] = param_value
            elif bead_names[0] in self.cross_library:
                if bead_names[1] in self.cross_library[bead_names[0]]:
                    self.cross_library[bead_names[0]][bead_names[1]][param_name] = param_value
                else:
                    self.cross_library[bead_names[0]][bead_names[1]] = {param_name: param_value}
            else:
                self.cross_library[bead_names[0]] = {bead_names[1]: {param_name: param_value}}



import numpy as np
import numba

from despasito.equations_of_state import constants


def calc_Xika(indices, rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc):
    r"""
    A wrapper to calculate the fraction of molecules of component i that are not
    bonded at a site of type a on group k. Switched between functions for different
    Kklab

    Parameters
    ----------
    indices : list[list]
        A list of sets of (component, bead, site) to identify the values of the Xika
        matrix that are being fit
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    molecular_composition : numpy.array
        :math:`\nu_{i,k}/k_B`, Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    nk : numpy.ndarray
        For each bead the number of each type of site
    Fklab : numpy.ndarray
        The association strength between a site of type a on a group of type k of
        component i and a site of type b on a group of type l of component j., known
        as the Mayer f-function.
    Kklab : numpy.ndarray
        Bonding volume between each association site
    gr_assoc : numpy.ndarray
        Reference fluid pair correlation function used in calculating association
        sites, (len(rho) x Ncomp x Ncomp)

    Returns
    -------
    Xika : numpy.ndarray
        The fraction of molecules of component i that are not bonded at a site of
        type a on group k. Matrix (len(rho) x Ncomp x Nbeads x len(sitenames))
    err_array : numpy.ndarray
        Of the same length of rho, is a list in the error of the total error Xika for
        each point.
    """

    l_K = len(np.shape(Kklab))

    # Ensure all inputs are numpy arrays
    tmp_array = [rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc]
    for i, tmp in enumerate(tmp_array):
        if np.shape(tmp):
            tmp_array[i] = np.array(tmp)
        else:
            tmp_array[i] = np.array([tmp])
    rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc = tmp_array

    if l_K == 4:
        Xika_final, err_array = calc_Xika_4(indices, rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc)
    if l_K == 6:
        Xika_final, err_array = calc_Xika_6(indices, rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc)

    return Xika_final, err_array


# @numba.njit(
#    numba.types.Tuple((numba.f8[:, :], numba.f8[:]))(
#        numba.i8[:, :],
#        numba.f8[:],
#        numba.f8[:],
#        numba.f8[:, :],
#        numba.i8[:, :],
#        numba.f8[:, :, :, :],
#        numba.f8[:, :, :, :],
#        numba.f8[:, :, :],
#    )
# )
@numba.njit()
def calc_Xika_4(
    indices, rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc
):  # , maxiter=500, tol=1e-12, damp=.1
    r"""
    Calculate the fraction of molecules of component i that are not bonded at a
    site of type a on group k.

    Parameters
    ----------
    indices : list[list]
        A list of sets of (component, bead, site) to identify the values of the Xika
        matrix that are being fit
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    molecular_composition : numpy.array
        :math:`\nu_{i,k}/k_B`, Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    nk : numpy.ndarray
        For each bead the number of each type of site
    Fklab : numpy.ndarray
        The association strength between a site of type a on a group of type k of
        component i and a site of type b on a group of type l of component j., known
        as the Mayer f-function.
    Kklab : numpy.ndarray
        Bonding volume between each association site
    gr_assoc : numpy.ndarray
        Reference fluid pair correlation function used in calculating association
        sites, (len(rho) x Ncomp x Ncomp)

    Returns
    -------
    Xika : numpy.ndarray
        The fraction of molecules of component i that are not bonded at a site of
        type a on group k. Matrix (len(rho) x Ncomp x Nbeads x len(sitenames))
    err_array : numpy.ndarray
        Of the same length of rho, is a list in the error of the total error Xika
        for each point.
    """

    maxiter = 500
    tol = 1e-12
    damp = 0.1

    nrho = len(rho)
    l_ind = len(indices)

    #    Xika_final = np.ones((nrho,ncomp, nbeads, nsitesmax))
    Xika_final = np.ones((nrho, l_ind))
    err_array = np.zeros(nrho)

    # Parallelize here, wrt rho!
    Xika_elements = 0.5 * np.ones(l_ind)
    for r in range(nrho):
        for knd in range(maxiter):

            Xika_elements_new = np.ones(l_ind)
            ind = 0
            for iind in range(l_ind):
                i, k, a = indices[iind]
                jnd = 0
                for jjnd in range(l_ind):
                    j, l, b = indices[jjnd]
                    delta = Fklab[k, l, a, b] * Kklab[k, l, a, b] * gr_assoc[r, i, j]
                    Xika_elements_new[ind] += (
                        constants.molecule_per_nm3
                        * rho[r]
                        * xi[j]
                        * molecular_composition[j, l]
                        * nk[l, b]
                        * Xika_elements[jnd]
                        * delta
                    )
                    jnd += 1
                ind += 1

            Xika_elements_new = 1.0 / Xika_elements_new
            obj = np.sum(np.abs(Xika_elements_new - Xika_elements))

            if obj < tol:
                break
            else:
                if obj / np.max(Xika_elements) > 1e3:
                    Xika_elements = Xika_elements + damp * (Xika_elements_new - Xika_elements)
                else:
                    Xika_elements = Xika_elements_new

        err_array[r] = obj
        Xika_final[r, :] = Xika_elements

    return Xika_final, err_array


# @numba.njit(
#    numba.types.Tuple((numba.f8[:, :], numba.f8[:]))(
#        numba.i8[:, :],
#        numba.f8[:],
#        numba.f8[:],
#        numba.f8[:, :],
#        numba.i8[:, :],
#        numba.f8[:, :, :, :],
#        numba.f8[:, :, :, :, :, :],
#        numba.f8[:, :, :],
#    )
# )
@numba.njit()
def calc_Xika_6(
    indices, rho, xi, molecular_composition, nk, Fklab, Kklab, gr_assoc
):  # , maxiter=500, tol=1e-12, damp=.1
    r"""
    Calculate the fraction of molecules of component i that are not bonded at a site
    of type a on group k.

    Parameters
    ----------
    indices : list[list]
        A list of sets of (component, bead, site) to identify the values of the Xika
        matrix that are being fit
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    molecular_composition : numpy.array
        :math:`\nu_{i,k}/k_B`, Array of number of components by number of bead types.
        Defines the number of each type of group in each component.
    nk : numpy.ndarray
        For each bead the number of each type of site
    Fklab : numpy.ndarray
        The association strength between a site of type a on a group of type k of
        component i and a site of type b on a group of type l of component j., known
        as the Mayer f-function.
    Kklab : numpy.ndarray
        Bonding volume between each association site
    gr_assoc : numpy.ndarray
        Reference fluid pair correlation function used in calculating association
        sites, (len(rho) x Ncomp x Ncomp)

    Returns
    -------
    Xika : numpy.ndarray
        The fraction of molecules of component i that are not bonded at a site of type
        a on group k. Matrix (len(rho) x Ncomp x Nbeads x len(sitenames))
    err_array : numpy.ndarray
        Of the same length of rho, is a list in the error of the total error Xika for
        each point.
    """

    maxiter = 500
    tol = 1e-12
    damp = 0.1

    nrho = len(rho)
    l_ind = len(indices)

    #    Xika_final = np.ones((nrho,ncomp, nbeads, nsitesmax))
    Xika_final = np.ones((nrho, l_ind))
    err_array = np.zeros(nrho)

    # Parallelize here, wrt rho!
    Xika_elements = 0.5 * np.ones(l_ind)
    for r in range(nrho):
        for knd in range(maxiter):

            Xika_elements_new = np.ones(l_ind)
            ind = 0
            for iind in range(l_ind):
                i, k, a = indices[iind]
                jnd = 0
                for jjnd in range(l_ind):
                    j, l, b = indices[jjnd]
                    delta = Fklab[k, l, a, b] * Kklab[i, j, k, l, a, b] * gr_assoc[r, i, j]
                    Xika_elements_new[ind] += (
                        constants.molecule_per_nm3
                        * rho[r]
                        * xi[j]
                        * molecular_composition[j, l]
                        * nk[l, b]
                        * Xika_elements[jnd]
                        * delta
                    )
                    jnd += 1
                ind += 1

            Xika_elements_new = 1.0 / Xika_elements_new
            obj = np.sum(np.abs(Xika_elements_new - Xika_elements))

            if obj < tol:
                break
            else:
                if obj / np.max(Xika_elements) > 1e3:
                    Xika_elements = Xika_elements + damp * (Xika_elements_new - Xika_elements)
                else:
                    Xika_elements = Xika_elements_new

        err_array[r] = obj
        Xika_final[r, :] = Xika_elements

    return Xika_final, err_array



r"""
    Routines for calculating the Helmholtz energy for the SAFT-gamma equation of state.

    Equations referenced in this code are from V. Papaioannou et al. J. Chem. Phys.
    140 054107 2014
"""

import numpy as np

from despasito.equations_of_state import constants

# from profilehooks import profile

# For Numba, ckl_coef cannot be encapsulated
ckl_coef = np.array(
    [
        [0.81096, 1.7888, -37.578, 92.284],
        [1.0205, -19.341, 151.26, -463.50],
        [-1.9057, 22.845, -228.14, 973.92],
        [1.0885, -6.1962, 106.98, -677.64],
    ]
)


def prefactor(l_r, l_a):
    r"""
    Calculations C, the Mie potential prefactor, defined in eq. 2

    Parameters
    ----------
    l_a : float
        Mie potential attractive exponent
    l_r : float
        Mie potential attractive exponent

    Returns
    -------
    C : float
        Mie potential prefactor
    """
    if np.any(l_r == 0.0):
        raise ValueError("l_r should not be zero.")

    return (l_r / (l_r - l_a)) * (l_r / l_a) ** (l_a / (l_r - l_a))


def calc_a1s(rho, Cmol2seg, l_kl, zetax, epsilonkl, dkl):
    r"""
    Return a1s,kl(rho*Cmol2seg,l_kl) in K as defined in eq. 25.

    Used in the calculation of :math:`A_1` the first order term of the perturbation
    expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_kl : numpy.ndarray
        Matrix of Mie potential exponents for k,l groups
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)

    Returns
    -------
    a1s : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the
        perturbation expansion corresponding to the mean-attractive energy, size is the
        Nrho x Ngroups x Ngroups

    """

    nbeads = len(dkl)
    zetax_pow = np.zeros((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1, 4):
        zetax_pow[:, i] = zetax_pow[:, i - 1] * zetax_pow[:, 0]

    etakl = np.zeros((len(rho), nbeads, nbeads), dtype=rho.dtype)

    for k in range(nbeads):
        for l in range(nbeads):
            tmp = np.dot(
                ckl_coef,
                np.array(
                    (
                        1.0,
                        1.0 / l_kl[k, l],
                        1.0 / l_kl[k, l] ** 2,
                        1.0 / l_kl[k, l] ** 3,
                    ),
                    dtype=ckl_coef.dtype,
                ),
            )
            etakl[:, k, l] = np.dot(zetax_pow, tmp)
    a1s = (
        -(1.0 - (etakl / 2.0))
        / ((1.0 - etakl) ** 3)
        * 2.0
        * np.pi
        * Cmol2seg
        * ((epsilonkl * (dkl**3 * constants.molecule_per_nm3**2)) / (l_kl - 3.0))
    )

    output = np.transpose(np.transpose(a1s) * rho)

    return output


def calc_Bkl(rho, l_kl, Cmol2seg, dkl, epsilonkl, x0kl, zetax):
    r"""
    Return Bkl(rho*Cmol2seg,l_kl) in K as defined in eq. 20.

    Used in the calculation of the first order term of the perturbation expansion
    corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    l_kl : numpy.ndarray
        :math:`\lambda_{k,l}` Matrix of Mie potential exponents for k,l groups
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    x0kl : numpy.ndarray
        Matrix of sigmakl/dkl, ratio of Mie radius for groups scaled by hard sphere
        interaction (k,l)
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)

    Returns
    -------
    Bkl : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the
        perturbation expansion corresponding to the mean-attractive energy, size is rho
        x l_kl.shape

    """

    rhos = Cmol2seg * rho

    # compute Ikl(l_kl), eq. 23
    Ikl = (1.0 - (x0kl ** (3.0 - l_kl))) / (l_kl - 3.0)
    # compute Jkl(l_kl), eq. 24
    Jkl = (1.0 - ((x0kl ** (4.0 - l_kl)) * (l_kl - 3.0)) + ((x0kl ** (3.0 - l_kl)) * (l_kl - 4.0))) / (
        (l_kl - 3.0) * (l_kl - 4.0)
    )

    tmp11 = rhos * (2.0 * np.pi)
    tmp12 = (dkl**3 * constants.molecule_per_nm3**2) * epsilonkl
    tmp2 = (1.0 - (zetax / 2.0)) / ((1.0 - zetax) ** 3)
    tmp3 = (9.0 * zetax * (1.0 + zetax)) / (2.0 * ((1 - zetax) ** 3))

    lx = len(Ikl)
    Bkl = np.zeros((len(rho), lx, lx))
    for i in np.arange(lx):
        for j in np.arange(lx):
            Bkl[:, i, j] = tmp11 * tmp12[i, j] * (tmp2 * Ikl[i, j] - tmp3 * Jkl[i, j])

    return Bkl


def calc_a1ii(rho, Cmol2seg, dkl, l_akl, l_rkl, x0kl, epsilonkl, zetax):
    r"""
    Calculate effective first-order perturbation term :math:`\bar{a}_{1,ii}`.

    Used for the contribution of the monomeric interactions to the free energy per
    segment.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dkl : numpy.ndarray
        Matrix of hard sphere diameters for groups (k,l)
    l_rkl : numpy.ndarray
        :math:`\lambda_{k,l}` Matrix of Mie potential repulsive exponents for
        k,l groups
    l_akl : numpy.ndarray
        :math:`\lambda_{k,l}` Matrix of Mie potential attractive exponents for
        k,l groups
    x0kl : numpy.ndarray
        Matrix of sigmakl/dkl, ratio of Mie radius for groups scaled by hard sphere
        interaction (k,l)
    epsilonkl : numpy.ndarray
        Matrix of well depths for groups (k,l)
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)

    Returns
    -------
    a1ii : numpy.ndarray
        Matrix used in the calculation of the radial distribution function of a
        hypothetical one-fluid Mie system.
    """
    Cii = (l_rkl / (l_rkl - l_akl)) * (l_rkl / l_akl) ** (l_akl / (l_rkl - l_akl))

    Bii_r = calc_Bkl(rho, l_rkl, Cmol2seg, dkl, epsilonkl, x0kl, zetax)
    Bii_a = calc_Bkl(rho, l_akl, Cmol2seg, dkl, epsilonkl, x0kl, zetax)
    a1s_r = calc_a1s(rho, Cmol2seg, l_rkl, zetax, epsilonkl, dkl)
    a1s_a = calc_a1s(rho, Cmol2seg, l_akl, zetax, epsilonkl, dkl)

    output = Cii * (
        ((x0kl**l_akl) / constants.molecule_per_nm3 * (a1s_a + Bii_a))
        - ((x0kl**l_rkl) / constants.molecule_per_nm3 * (a1s_r + Bii_r))
    )

    return output


def calc_a1s_eff(rho, Cmol2seg, l_ii_avg, zetax, epsilonii_avg, dii_avg):
    r"""
    Return a1s,(rho*Cmol2seg,l_ij) in K

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to s
        egment (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_ii_avg : numpy.ndarray
        Average bead (i.e. group or segment) exponent in component (i.e. molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component
        (i.e. molecule) i.
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in
        component (i.e. molecule) i.

    Returns
    -------
    a1s_eff : numpy.ndarray
        Term used in the used in the calculation of the effective averaged molecular
        radial distribution function of a hypothetical one-fluid Mie system.
    """

    ncomp = len(dii_avg)
    zetax_pow = np.zeros((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1, 4):
        zetax_pow[:, i] = zetax_pow[:, i - 1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    etaii_avg = np.zeros((len(rho), ncomp), dtype=rho.dtype)

    for k in range(ncomp):
        tmp = np.dot(
            ckl_coef,
            np.array(
                (
                    1.0,
                    1.0 / l_ii_avg[k],
                    1.0 / l_ii_avg[k] ** 2,
                    1.0 / l_ii_avg[k] ** 3,
                ),
                dtype=ckl_coef.dtype,
            ),
        )
        etaii_avg[:, k] = np.dot(zetax_pow, tmp)

    a1s = (
        -(1.0 - (etaii_avg / 2.0))
        / (1.0 - etaii_avg) ** 3
        * 2.0
        * np.pi
        * Cmol2seg
        * ((epsilonii_avg * (dii_avg**3 * constants.molecule_per_nm3**2)) / (l_ii_avg - 3.0))
    )

    output = np.transpose(np.transpose(a1s) * rho)

    return output


def calc_Bkl_eff(rho, l_ii_avg, Cmol2seg, dii_avg, epsilonii_avg, x0ii, zetax):
    r"""
    Return Bii_avg(rho*Cmol2seg,l_ii_avg) in K as defined in eq. 20.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    l_ii_avg : numpy.ndarray
        Average bead (i.e. group or segment) exponent in component (i.e. molecule) i.
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in
        component (i.e. molecule) i.
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e.
        molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)

    Returns
    -------
    Bii_avg : numpy.ndarray
        Bii_avg(rho*Cmol2seg,l_ii_avg) in K as defined in eq. 20.

    """

    rhos = Cmol2seg * rho
    ncomp = len(dii_avg)

    # compute Iii_avg(l_ii_avg), eq. 23
    Iii_avg = (1.0 - (x0ii ** (3.0 - l_ii_avg))) / (l_ii_avg - 3.0)
    # compute Jii_avg(l_ii_avg), eq. 24
    Jii_avg = (
        1.0 - ((x0ii ** (4.0 - l_ii_avg)) * (l_ii_avg - 3.0)) + ((x0ii ** (3.0 - l_ii_avg)) * (l_ii_avg - 4.0))
    ) / ((l_ii_avg - 3.0) * (l_ii_avg - 4.0))

    tmp11 = rhos * (2.0 * np.pi)
    tmp12 = (dii_avg**3 * constants.molecule_per_nm3**2) * epsilonii_avg
    tmp2 = (1.0 - (zetax / 2.0)) / ((1.0 - zetax) ** 3)
    tmp3 = (9.0 * zetax * (1.0 + zetax)) / (2.0 * ((1 - zetax) ** 3))

    Bii_avg = np.zeros((len(rho), ncomp))
    for k in np.arange(ncomp):
        Bii_avg[:, k] = tmp11 * tmp12[k] * (tmp2 * Iii_avg[k] - tmp3 * Jii_avg[k])
    #  Bii_avg = tmp11*tmp12*(tmp2*Iii_avg - tmp3*Jii_avg)

    return Bii_avg


def calc_da1sii_drhos(rho, Cmol2seg, l_ii_avg, zetax, epsilonii_avg, dii_avg):
    r"""
    Return the derivative of a1s,ii_avg(rho*Cmol2seg,l_ii_avg) with represent to
    number density in Kelvin

    Used in the calculation of the first order term of the perturbation expansion
    corresponding to the mean-attractive energy.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    l_ii_avg : numpy.ndarray
        Average bead (i.e. group or segment) exponent in component (i.e. molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e.
        molecule) i.
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in
        component (i.e. molecule) i.

    Returns
    -------
    calc_da1sii_drhos : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the
        perturbation expansion corresponding to the mean-attractive energy
    """
    ncomp = len(dii_avg)
    zetax_pow = np.zeros((len(rho), 4), dtype=rho.dtype)
    zetax_pow[:, 0] = zetax
    for i in range(1, 4):
        zetax_pow[:, i] = zetax_pow[:, i - 1] * zetax_pow[:, 0]

    # check if you have more than 1 bead types
    etaii_avg = np.zeros((len(rho), ncomp), dtype=rho.dtype)
    rhos_detaii_avg_drhos = np.zeros((len(rho), ncomp), dtype=rho.dtype)

    for k in range(ncomp):
        ciii_avg = np.dot(
            ckl_coef,
            np.array(
                (
                    1.0,
                    1.0 / l_ii_avg[k],
                    1.0 / l_ii_avg[k] ** 2,
                    1.0 / l_ii_avg[k] ** 3,
                ),
                dtype=ckl_coef.dtype,
            ),
        )
        etaii_avg[:, k] = np.dot(zetax_pow, ciii_avg)
        rhos_detaii_avg_drhos[:, k] = np.dot(zetax_pow, ciii_avg * np.array([1.0, 2.0, 3.0, 4.0]))

    tmp1 = (1.0 - (etaii_avg / 2.0)) / ((1.0 - etaii_avg) ** 3) + (5.0 - 2.0 * etaii_avg) / (
        2.0 * (1.0 - etaii_avg) ** 4
    ) * rhos_detaii_avg_drhos
    tmp2 = -2.0 * np.pi * ((epsilonii_avg * (dii_avg**3 * constants.molecule_per_nm3**2)) / (l_ii_avg - 3.0))

    da1s_drhos = tmp1 * tmp2

    return da1s_drhos


def calc_dBkl_drhos(l_ii_avg, dii_avg, epsilonii_avg, x0ii, zetax):
    r"""
    Return derivative of Bkl(rho*Cmol2seg,l_ii_avg) with respect to :math:`\rho_S`.

    Used in the calculation of :math:`A_1` the first order term of the perturbation
    expansion corresponding to the mean-attractive energy.

    Parameters
    ----------
    l_aii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e.
        molecule) i.
    dii_avg : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in
        component (i.e. molecule) i.
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e.
        molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)

    Returns
    -------
    dBkl_drhos : numpy.ndarray
        Matrix used in the calculation of :math:`A_1` the first order term of the
        perturbation expansion corresponding to the mean-attractive energy, size is
        rho x l_ii_avg.shape

    """
    ncomp = len(dii_avg)
    nrho = len(zetax)

    # compute Iii_avg(l_ii_avg), eq. 23
    Iii_avg = (1.0 - (x0ii ** (3.0 - l_ii_avg))) / (l_ii_avg - 3.0)
    # compute Jii_avg(l_ii_avg), eq. 24
    Jii_avg = (
        1.0 - ((x0ii ** (4.0 - l_ii_avg)) * (l_ii_avg - 3.0)) + ((x0ii ** (3.0 - l_ii_avg)) * (l_ii_avg - 4.0))
    ) / ((l_ii_avg - 3.0) * (l_ii_avg - 4.0))

    tmp = 2.0 * np.pi * dii_avg**3 * epsilonii_avg
    tmp1 = np.zeros((nrho, ncomp))
    tmp2 = np.zeros((nrho, ncomp))
    for k in np.arange(ncomp):
        tmp1[:, k] = ((1.0 - (zetax / 2.0)) / ((1.0 - zetax) ** 3) * Iii_avg[k]) - (
            ((9.0 * zetax * (1.0 + zetax)) / (2.0 * ((1 - zetax) ** 3))) * Jii_avg[k]
        )
        tmp2[:, k] = ((5.0 - 2.0 * zetax) * zetax / (2 * (1.0 - zetax) ** 4) * Iii_avg[k]) - (
            ((9.0 * zetax * (zetax**2 + 4.0 * zetax + 1)) / (2.0 * ((1 - zetax) ** 4))) * Jii_avg[k]
        )
    dBkl_drhos = tmp * (tmp1 + tmp2) * constants.molecule_per_nm3**2

    return dBkl_drhos


def calc_da1iidrhos(rho, Cmol2seg, dii_eff, l_aii_avg, l_rii_avg, x0ii, epsilonii_avg, zetax):
    r"""
    Compute derivative of the term,
    :math:`\bar{a}_{1,ii}` with respect to :math:`\rho_s`

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    dii_eff : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in
        component (i.e. molecule) i.
    l_aii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e.
        molecule) i.
    l_rii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e.
        molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e.
        molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)

    Returns
    -------
    da1iidrhos : numpy.ndarray
        Derivative of term with respect to segment density
    """

    Cii = prefactor(l_rii_avg, l_aii_avg)

    das1_drhos_a = calc_da1sii_drhos(rho, Cmol2seg, l_aii_avg, zetax, epsilonii_avg, dii_eff)
    das1_drhos_r = calc_da1sii_drhos(rho, Cmol2seg, l_rii_avg, zetax, epsilonii_avg, dii_eff)

    dB_drhos_a = calc_dBkl_drhos(l_aii_avg, dii_eff, epsilonii_avg, x0ii, zetax)
    dB_drhos_r = calc_dBkl_drhos(l_rii_avg, dii_eff, epsilonii_avg, x0ii, zetax)

    da1iidrhos = Cii * (
        ((x0ii**l_aii_avg) * (das1_drhos_a + dB_drhos_a)) - ((x0ii**l_rii_avg) * (das1_drhos_r + dB_drhos_r))
    )

    return da1iidrhos


def calc_da2ii_1pchi_drhos(rho, Cmol2seg, epsilonii_avg, dii_eff, x0ii, l_rii_avg, l_aii_avg, zetax):
    r"""
    Compute derivative of the term, :math:`\frac{\bar{a}_{2,ii}}{1+\bar{\chi}_{ii}}`
    with respect to :math:`\rho_s`.

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    Cmol2seg : float
        Conversion factor from from molecular number density, :math:`\rho`, to segment
        (i.e. group) number density, :math:`\rho_S`. Shown in eq. 13
    epsilonii_avg : numpy.ndarray
        Average bead (i.e. group or segment) potential well depth in component (i.e.
        molecule) i.
    dii_eff : numpy.ndarray
        Effective hard sphere diameter of the beads (i.e. groups or segments) in
        component (i.e. molecule) i.
    x0ii : numpy.ndarray
        Matrix of sigmaii_avg/dii_eff
    l_rii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e.
        molecule) i.
    l_aii_avg : numpy.ndarray
        Average bead (i.e. group or segment) attractive exponent in component (i.e.
        molecule) i.
    zetax : numpy.ndarray
        Matrix of hypothetical packing fraction based on hard sphere diameter for
        groups (k,l)

    Returns
    -------
    da2ii_1pchi_drhos : numpy.ndarray
        Term used in the calculation of the second-order term from the macroscopic
        compressibility

    """

    # Calculate terms and derivatives used in derivative chain rule
    KHS = ((1.0 - zetax) ** 4) / (1.0 + (4.0 * zetax) + (4.0 * (zetax**2)) - (4.0 * (zetax**3)) + (zetax**4))
    dKHS_drhos = (
        (4.0 * (zetax**2 - 5.0 * zetax - 2.0) * (1.0 - zetax) ** 3)
        / (zetax**4 - 4.0 * zetax**3 + 4.0 * zetax**2 + 4.0 * zetax + 1.0) ** 2
        * (zetax / (rho * Cmol2seg))
    )

    a1sii_2l_aii_avg = calc_a1s_eff(rho, Cmol2seg, 2.0 * l_aii_avg, zetax, epsilonii_avg, dii_eff)
    a1sii_2l_rii_avg = calc_a1s_eff(rho, Cmol2seg, 2.0 * l_rii_avg, zetax, epsilonii_avg, dii_eff)
    a1sii_l_rii_avgl_aii_avg = calc_a1s_eff(rho, Cmol2seg, l_aii_avg + l_rii_avg, zetax, epsilonii_avg, dii_eff)

    Bii_2l_aii_avg = calc_Bkl_eff(rho, 2.0 * l_aii_avg, Cmol2seg, dii_eff, epsilonii_avg, x0ii, zetax)
    Bii_2l_rii_avg = calc_Bkl_eff(rho, 2.0 * l_rii_avg, Cmol2seg, dii_eff, epsilonii_avg, x0ii, zetax)
    Bii_l_aii_avgl_rii_avg = calc_Bkl_eff(rho, l_aii_avg + l_rii_avg, Cmol2seg, dii_eff, epsilonii_avg, x0ii, zetax)

    da1sii_2l_aii_avg = calc_da1sii_drhos(rho, Cmol2seg, 2.0 * l_aii_avg, zetax, epsilonii_avg, dii_eff)
    da1sii_2l_rii_avg = calc_da1sii_drhos(rho, Cmol2seg, 2.0 * l_rii_avg, zetax, epsilonii_avg, dii_eff)
    da1sii_l_rii_avgl_aii_avg = calc_da1sii_drhos(rho, Cmol2seg, l_aii_avg + l_rii_avg, zetax, epsilonii_avg, dii_eff)

    dBii_2l_aii_avg = calc_dBkl_drhos(2.0 * l_aii_avg, dii_eff, epsilonii_avg, x0ii, zetax)
    dBii_2l_rii_avg = calc_dBkl_drhos(2.0 * l_rii_avg, dii_eff, epsilonii_avg, x0ii, zetax)
    dBii_l_aii_avgl_rii_avg = calc_dBkl_drhos(l_aii_avg + l_rii_avg, dii_eff, epsilonii_avg, x0ii, zetax)

    # Calculate Derivative
    Cii = prefactor(l_rii_avg, l_aii_avg)

    B = (
        x0ii ** (2.0 * l_aii_avg) * (a1sii_2l_aii_avg + Bii_2l_aii_avg)
        - 2.0 * x0ii ** (l_aii_avg + l_rii_avg) * (a1sii_l_rii_avgl_aii_avg + Bii_l_aii_avgl_rii_avg)
        + x0ii ** (2.0 * l_rii_avg) * (a1sii_2l_rii_avg + Bii_2l_rii_avg)
    )

    dA_B = np.transpose(np.transpose(0.5 * epsilonii_avg * Cii**2 * B) * dKHS_drhos)

    dB = (
        x0ii ** (2.0 * l_aii_avg) * (da1sii_2l_aii_avg + dBii_2l_aii_avg)
        - 2.0 * x0ii ** (l_aii_avg + l_rii_avg) * (da1sii_l_rii_avgl_aii_avg + dBii_l_aii_avgl_rii_avg)
        + x0ii ** (2.0 * l_rii_avg) * (da1sii_2l_rii_avg + dBii_2l_rii_avg)
    )

    A_dB = np.transpose(np.transpose(0.5 * epsilonii_avg * Cii**2 * dB) * KHS)

    da2ii_1pchi_drhos = A_dB + dA_B

    return da2ii_1pchi_drhos


def calc_Iij(rho, T, xi, epsilonii_avg, sigmaii_avg, sigmakl, xskl):
    r"""

    Reference fluid pair correlation function used in calculating association sites

    Parameters
    ----------
    rho : numpy.ndarray
        Number density of system [mol/m^3]
    T : float
        Temperature of the system [K]
    xi : numpy.ndarray
        Mole fraction of each component, sum(xi) should equal 1.0
    epsilonii_avg : numpy.ndarray
        Array of average Mie diameters for component i
    sigmaii_avg : numpy.ndarray
        Array of average Mie diameters for component i
    sigmakl : numpy.ndarray
        Matrix of Mie diameter for groups (k,l)
    xskl : numpy.ndarray
        Matrix of mole fractions of bead (i.e. segment or group) k multiplied by that
        of bead l

    Returns
    -------
    Iij : numpy.ndarray
        A temperature-density polynomial correlation of the association integral for a
        Lennard-Jones monomer. This matrix is (len(rho) x Ncomp x Ncomp)

    """

    ncomp = len(xi)

    # compute epsilonij
    epsilonij = np.zeros((ncomp, ncomp))
    for i in range(ncomp):
        for j in range(i, ncomp):
            epsilonij[i, j] = (
                np.sqrt(sigmaii_avg[i] * sigmaii_avg[j]) ** 3.0
                * np.sqrt(epsilonii_avg[i] * epsilonii_avg[j])
                / (((sigmaii_avg[i] + sigmaii_avg[j]) / 2.0) ** 3)
            )
            epsilonij[j, i] = epsilonij[i, j]

    sigmax3 = np.sum(xskl * (sigmakl**3 * constants.molecule_per_nm3))

    cij = np.array(
        [
            [
                7.56425183020431e-02,
                -1.28667137050961e-01,
                1.28350632316055e-01,
                -7.25321780970292e-02,
                2.57782547511452e-02,
                -6.01170055221687e-03,
                9.33363147191978e-04,
                -9.55607377143667e-05,
                6.19576039900837e-06,
                -2.30466608213628e-07,
                3.74605718435540e-09,
            ],
            [
                1.34228218276565e-01,
                -1.82682168504886e-01,
                7.71662412959262e-02,
                -7.17458641164565e-04,
                -8.72427344283170e-03,
                2.97971836051287e-03,
                -4.84863997651451e-04,
                4.35262491516424e-05,
                -2.07789181640066e-06,
                4.13749349344802e-08,
                0.00000000000000e00,
            ],
            [
                -5.65116428942893e-01,
                1.00930692226792e00,
                -6.60166945915607e-01,
                2.14492212294301e-01,
                -3.88462990166792e-02,
                4.06016982985030e-03,
                -2.39515566373142e-04,
                7.25488368831468e-06,
                -8.58904640281928e-08,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                -3.87336382687019e-01,
                -2.11614570109503e-01,
                4.50442894490509e-01,
                -1.76931752538907e-01,
                3.17171522104923e-02,
                -2.91368915845693e-03,
                1.30193710011706e-04,
                -2.14505500786531e-06,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                2.13713180911797e00,
                -2.02798460133021e00,
                3.36709255682693e-01,
                1.18106507393722e-03,
                -6.00058423301506e-03,
                6.26343952584415e-04,
                -2.03636395699819e-05,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                -3.00527494795524e-01,
                2.89920714512243e00,
                -5.67134839686498e-01,
                5.18085125423494e-02,
                -2.39326776760414e-03,
                4.15107362643844e-05,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                -6.21028065719194e00,
                -1.92883360342573e00,
                2.84109761066570e-01,
                -1.57606767372364e-02,
                3.68599073256615e-04,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                1.16083532818029e01,
                7.42215544511197e-01,
                -8.23976531246117e-02,
                1.86167650098254e-03,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                -1.02632535542427e01,
                -1.25035689035085e-01,
                1.14299144831867e-02,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                4.65297446837297e00,
                -1.92518067137033e-03,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
            [
                -8.67296219639940e-01,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
                0.00000000000000e00,
            ],
        ]
    )

    Iij = np.zeros((np.size(rho), ncomp, ncomp))
    for p in range(11):
        for q in range(11 - p):
            # Iij += np.einsum("i,jk->ijk", cij[p, q] * ((sigmax3 * rho)**p),
            # ((T / epsilonij)**q))
            if p == 0:
                Iij += np.einsum("i,jk->ijk", cij[p, q] * np.ones(len(rho)), ((T / epsilonij) ** q))
            elif p == 1:
                Iij += np.einsum("i,jk->ijk", cij[p, q] * ((sigmax3 * rho)), ((T / epsilonij) ** q))
            elif p == 2:
                rho2 = rho**2
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho2)),
                    ((T / epsilonij) ** q),
                )
            elif p == 3:
                rho3 = rho2 * rho
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho3)),
                    ((T / epsilonij) ** q),
                )
            elif p == 4:
                rho4 = rho2**2
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho4)),
                    ((T / epsilonij) ** q),
                )
            elif p == 5:
                rho5 = rho2 * rho3
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho5)),
                    ((T / epsilonij) ** q),
                )
            elif p == 6:
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho * rho5)),
                    ((T / epsilonij) ** q),
                )
            elif p == 7:
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho2 * rho5)),
                    ((T / epsilonij) ** q),
                )
            elif p == 8:
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho3 * rho5)),
                    ((T / epsilonij) ** q),
                )
            elif p == 9:
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho4 * rho5)),
                    ((T / epsilonij) ** q),
                )
            elif p == 10:
                Iij += np.einsum(
                    "i,jk->ijk",
                    cij[p, q] * ((sigmax3**p * rho5 * rho5)),
                    ((T / epsilonij) ** q),
                )

    return Iij
